#include <SDL.h>
#include <SDL_thread.h>
#include <SDL_image.h>


#include <stdexcept>
#include <deque>
using namespace std;

#include "Graphs.h"
#include "Widget.h"
#include "Application.h"
#include "InputMap.h"
#include "JTimer.h"
#include "ImageLoader.h"
#include "Processes.h"
#include "Database.h"
#include "lirc.h"
#include "Task.h"
#include "JWidgetClass.h"
#include "Metadata.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <errno.h>
#include <fcntl.h> 

Application * Application::current = 0;
char * Application::slang_tmp_string = 0;
int Application::slang_tmp_int = 0;

static deque<SDL_Event> events;
static SDL_mutex * eventMutex;
static SDL_Thread * mainThread;
static int pipe_wr,pipe_rd;

bool CompareTimerClass::operator() (const Timer * a,const Timer * b) const 
{
    if ((!a->hasTimeout) && (!b->hasTimeout)) {
	return false;
    }
    
    if (!b->hasTimeout) {
	return true;
    }
    
    return a->timeoutAt < b->timeoutAt;
}

Application::Application():programmedEvents()
{
    inputDisabled = 0;
    lastNow = 0;
    dayInc = 0;
    current = this;
    simulateFullscreen = true;
    if (getenv("OSTBOX_FULLSCREEN")) {
	simulateFullscreen = false;
    }
    
    JSRuntime * js_runtime;

    /* initialize the JS run time, and return result in rt */
    js_runtime = JS_NewRuntime( 64L * 1024L /* * 1024L */ );

    /* if rt does not have a value, end the program here */
    if ( !js_runtime )
        throw 1;
    
    inputMap = new InputMap();
    
    jsContext = new JSReflection::Context(js_runtime,65536);
    deque<JSReflection::ClassInfo *> bidouf;
    bidouf.push_back(JTimer::ClassInfo);
    bidouf.push_back(LircListener::ClassInfo);
    bidouf.push_back(JSNotifiable::ClassInfo);
    bidouf.push_back(ImageLoader::ClassInfo);
    
    int pipefds[2];
    if (pipe(pipefds) == -1) {
	perror("pipe");
	throw runtime_error("pipe error");
    }
    
    pipe_rd = pipefds[0];
    pipe_wr = pipefds[1];
    fcntl(pipe_rd,F_SETFL,O_NONBLOCK);
    fcntl(pipe_wr,F_SETFL,O_NONBLOCK);
    
    InternalTask::initSyncThread(pipe_wr);
    InternalTask::startAsyncThreads(1);
    Database::initDatabase();
    WidgetClass::initialise();
}

InputMap * Application::getInputMap() const
{
    return inputMap;
}

JSReflection::Context * Application::getJsContext()
{
    return current->jsContext;
}

void Application::inputDisable(int n)
{
    inputDisabled += n;
}

/**
 * Return times in milliseconds.
 * times is not updated for each call, only when update is true.
 * First call always returns 1
 * 
 * NB : au moins sur sunos, gettimeofday retourne le nombre de secondes écoulées depuis... le 1er janvier 1970.
 * Celà implique que la multiplication par 1000 de cette valeur dépasse systématiquement la capacité d'un int 32bits,
 * et repasse à 0 tous les 50 jours.
 */
long long Application::now(bool update) {
    if (!update) {
	return lastNow;
    }
    
    struct timeval tv;
    long long nnow;
    if (gettimeofday(&tv,0) == -1) {
	perror("gettimeofday");
	/// Should we bother more ?
	return lastNow;
    }
    
    nnow = tv.tv_usec / 1000+ ((long long)tv.tv_sec)*1000;
    if (lastNow == 0) {
	srand(nnow);
	dayInc = (- nnow) + 1;
    }
    nnow += dayInc;
    if (nnow < lastNow) {
	jsrlog("lastNow was %lld, now is %lld\n",lastNow,nnow);
	while(nnow + 2000 < lastNow) {
	    // One more day !
	    nnow += 86400000;
	    dayInc += 86400000;
	    fprintf(stderr,"One more day elasped !\n");
	}
	if (nnow < lastNow) {
	    fprintf(stderr,"clock back detected.\n");
	    nnow = lastNow;
	}
    }
    lastNow = nnow;
    return nnow;
}

void Application::changeVideoMode(int w,int h)
{
    isFullscreen = true;

    screen=SDL_SetVideoMode(w,h,24,
		  (!simulateFullscreen?SDL_FULLSCREEN:SDL_ANYFORMAT)/*|SDL_HWSURFACE|SDL_DOUBLEBUF|*/
		  );
    if (screen==NULL) {
	char buffer[512];
	snprintf(buffer,512,"SetVideoMode failed : %s\n",SDL_GetError());
    	throw runtime_error(string(buffer));
    }
    
    SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY,SDL_DEFAULT_REPEAT_INTERVAL);
    SDL_ShowCursor(SDL_DISABLE);
    jsrlog("Video mode : %dx%dx%dbpp, hw=%d fullscreen=%d\n",
	      screen->w,screen->h,screen->format->BitsPerPixel,
	      (screen->flags&SDL_HWSURFACE)!=0,
	      (screen->flags&SDL_FULLSCREEN)!=0);
    screenSurface->assign(screen);
}

void Application::initialise(int w,int h)
{
    
    screenSurface = new Surface();
    changeVideoMode(w,h);
    
    
    /* Clean up on exit */
    atexit(SDL_Quit);
}

void Application::hide()
{
    jsrlog("hiding!\n");
    SDL_FreeSurface(screen);
    screenSurface->assign(0);
    jsrlog("done.\n");
}

void Application::show()
{
    changeVideoMode(800,600);
    screenSurface->assign(0);
}


void Application::setRoot(Widget * w)
{
    root = w;
}

Widget * Application::getRoot() const
{
    return root;
}

void Application::redraw()
{
    if ((!screenSurface) || (!root)) {
	return;
    }
    ScreenArea fullscreen(0,0,screenSurface->GetWidth(),screenSurface->GetHeight());
    root->draw(screenSurface,&fullscreen,0,0);
    SDL_Flip(screen);
    //updater.ask(false);
}

void Application::setFullscreen(bool b)
{
    if (isFullscreen == b) return;
    if (simulateFullscreen) {
	isFullscreen = b;
	return;
    }
    jsrlog("setting full screen to %d\n",(int)b);
    SDL_WM_GrabInput(b?SDL_GRAB_ON:SDL_GRAB_OFF);
    SDL_WM_ToggleFullScreen(screen);
    jsrlog("done\n");
    isFullscreen = b;
}



extern "C" {
static int application_run(void * v);
};

static bool takeEvent(SDL_Event & e)
{
    bool ret;
    SDL_mutexP(eventMutex);
    if (events.empty()) {
	ret = false;
    } else {
	e = events.front();
	events.pop_front();
	ret = true;
    }
    SDL_mutexV(eventMutex);
    return ret;
}

static int application_run(void *v)
{
    Application * a = (Application*)v;
    a->dorun();
    SDL_Event quit_event;
    
    quit_event.type=SDL_USEREVENT;
    SDL_PushEvent(&quit_event);
    
    return 0;
}

void Application::run()
{
    SDL_Event event;
    
    ChildProcess::installSigHandler();
    
    eventMutex = SDL_CreateMutex();
    mainThread = SDL_CreateThread(&application_run,this);
    ChildProcess::unmaskSignal();
    
    while(SDL_WaitEvent(&event)) {
	ChildProcess::maskSignal();
	
	if (event.type == SDL_USEREVENT) {
	    jsrlog("end event received...\n");
	    SDL_WaitThread(mainThread,0);
	    return;
	}
	SDL_mutexP(eventMutex);
	events.push_back(event);
	/// make sure the mainThread got the event !
	if (write(pipe_wr,&pipe_wr,1) == -1) {
	    perror("write into pipe");
	    break;
	}
	SDL_mutexV(eventMutex);
	ChildProcess::unmaskSignal();
    }
    SDL_KillThread(mainThread);
}

void Application::inputEvent(const std::string & tmp)
{
    if (inputDisabled) return;
    jsrlog("handle %s\n",tmp.c_str());
    if (!root->handleInputEvent(tmp,true)) {
	root->handleInputEvent(tmp,false);
    }
}

void Application::reportFileChanged(const std::string & n)
{
    jsrlog("file change detected on %s\n",n.c_str());
    InternalTask::lock();
    
    ImageFactory::invalidateImage(n);
    MetaUser::invalidateFile(n);
    
    InternalTask::unlock();
}

static bool terminate_flag = false;

static int quit()
{
    terminate_flag = true;
    return 1;
}

void Application::dorun()
{
    char buffer[256];
    long long n;
    n = now(true);
    updater.ask(true);
    terminate_flag = false;
    do {
	int sr,maxr,maxw;
	fd_set readset,writeset;
	
	maxr = -1;
	maxw = -1;
	FD_ZERO(&readset);      /* empties the set */
	FD_ZERO(&writeset);      /* empties the set */
	
	maxr = pipe_rd;
	FD_SET(pipe_rd,&readset);
	
	/// Find the lowest timeout
	bool hasTimeout = false;
	long long nextTimeout;
	
	for(set<Timer*>::const_iterator it=programmedEvents.begin(); it != programmedEvents.end(); ++it) {
	    if ((!hasTimeout) || (nextTimeout > (*it)->timeoutAt)) {
		hasTimeout = true;
		nextTimeout = (*it)->timeoutAt;
	    }
	}
	
	for(set<Selectable*>::const_iterator it=selectables.begin(); it != selectables.end(); ++it) {
	    if ((*it)->getFd() != -1) {
		FD_SET((*it)->getFd(),&readset);
		if ((*it)->getFd() > maxr) {
		    maxr = (*it)->getFd();
		}
	    }
	}
	
	/// select...
	do {
	    
	    struct timeval tv;
	    long long to;
	    if (hasTimeout) {
		to = now(true);
		if (nextTimeout > to) {
		    to = nextTimeout - to;
		} else {
		    to = 0;
		}
	    } else {
		to = 1000;
	    }
	    
	    tv.tv_usec = 1000*(to % 1000);
	    tv.tv_sec = to / 1000;
	    
	    sr = select(max(maxr,maxw) + 1,&readset,&writeset,0,&tv);
	} while((sr == -1) && (errno == EAGAIN));
	
	if (sr == -1) {
	    perror("select");
	    throw runtime_error("select failed");
	}
	
	n = now(true);
	
	/// Just kill pipe garbage...
	read(pipe_rd,&buffer,256);
	
	SDL_Event evt;
	
	/// Check for new events...
	while(takeEvent(evt)) {
	    if (evt.type == SDL_QUIT) {
		return;
	    }
	    if (evt.type==SDL_KEYDOWN) {
		std::string tmp(SDL_GetKeyName(evt.key.keysym.sym));
		inputEvent("KEY:"+tmp);
	    }
	    /*InputEvent ievt(evt);
	    root->handleInputEvent(ievt);*/
	}
	
	while(InternalTask::syncTaskQueue->step()) {
	}
	
	set<Selectable*> tmp = selectables;
	for(set<Selectable*>::const_iterator it=tmp.begin(); it != tmp.end(); ++it) {
	    if ((*it)->getFd() != -1) {
		if (FD_ISSET((*it)->getFd(),&readset)) {
		    (*it)->doRead();
		}
	    }
	}
	
	/// Iterate all timeouts
	do {
	    long long selectedTime;
	    Timer * selected = 0;
	    
	    for(set<Timer*>::const_iterator it=programmedEvents.begin(); it != programmedEvents.end(); ++it) {
		if (((*it)->timeoutAt <= n) && ((!selected) || selectedTime> (*it)->timeoutAt)) {
		    selected = *it;
		    selectedTime = (*it)->timeoutAt;
		}
	    }
	    
	    if (!selected) {
		break;
	    }
	    selected->hasTimeout = false;
	    programmedEvents.erase(selected);
	    selected->tick();
	}while(1);
	
    } while(!terminate_flag);
}

static Widget * loadXmlWidget(std::string name, Widget * father)
{
    return Application::loadXmlWidget(name.c_str(),father);
}

static int setFullscreen(bool on) {
    Application::current->setFullscreen(on);
    return 0;
}

static bool isFullscreen() {
    return Application::current->isFullscreen;
}

static int showMain() {
    Application::current->show();
    return 0;
}

static int hideMain() {
    Application::current->hide();
    return 0;
}

static int inputDisable() {
    Application::current->inputDisable(1);
    return 0;
}

static int inputEnable() {
    Application::current->inputDisable(-1);
    return 0;
}

static int fileChanged(const std::string & s)
{
    Application::reportFileChanged(s);
    return 0;
}

static std::string pathToDisplay(const std::string & s)
{
    return JSReflection::displayBinaryPath(s);
}

JS_NOCONTEXT_GLOBAL_FUNCTION_0(quit,int,"%i",0)
JS_NOCONTEXT_GLOBAL_FUNCTION_1(setFullscreen,int,"%i",0,bool,"%b",0)
JS_NOCONTEXT_GLOBAL_FUNCTION_0(isFullscreen,bool,"%b",0)
JS_NOCONTEXT_GLOBAL_FUNCTION_0(showMain,int,"%i",0)
JS_NOCONTEXT_GLOBAL_FUNCTION_0(hideMain,int,"%i",0)
JS_NOCONTEXT_GLOBAL_FUNCTION_0(inputEnable,int,"%i",0)
JS_NOCONTEXT_GLOBAL_FUNCTION_0(inputDisable,int,"%i",0)
JS_NOCONTEXT_GLOBAL_FUNCTION_2(loadXmlWidget,Widget*,"%x",Widget::ClassInfo,JSReflection::JS_String,"%s",0,Widget*,"%x",Widget::ClassInfo)
JS_NOCONTEXT_GLOBAL_FUNCTION_1(fileChanged,int,"%i",0,JSReflection::JS_String,"%s",0)
JS_NOCONTEXT_GLOBAL_FUNCTION_1(pathToDisplay,JSReflection::JS_String,"%s",0,JSReflection::JS_String,"%s",0)
