#include "jsreflect.h"

#include "Database.h"

Database::Database(const std::string & p) : path(p)
{
    ppDb = 0;
}

Database::~Database()
{
    if (ppDb) {
	close();
    }
}

bool Database::open()
{
    int rc;
    rc = sqlite3_open(path.c_str(),&ppDb);
    if( rc ){
	jsrlog("%s: Can't open database: %s\n", path.c_str(), sqlite3_errmsg(ppDb));
	sqlite3_close(ppDb);
	ppDb = 0;
	return false;
    }

    exec("PRAGMA auto_vacuum=1;");
    return true;
}

void Database::close()
{
    sqlite3_close(ppDb);
    ppDb = 0;
}

static int callback(void * db,int argc,char ** argv,char ** azColName)
{
    return 0;
}

bool Database::exec(const std::string & str)
{
    char * zErrMsg;
    int rc = sqlite3_exec(ppDb, str.c_str(), callback, (void*)this, &zErrMsg);
    if( rc != SQLITE_OK ){
	jsrlog("%s: SQL error: %s\n", path.c_str(), zErrMsg);
	return false;
    } else {
	return true;
    }
}

bool Database::exec(	const std::string & str,
			void (*data)(void *, std::vector<std::string*> & ),void * bindarg)
{
    std::map<std::string,std::string> noargs;
    return exec(str,noargs,data,bindarg);
}

bool Database::exec(	const std::string & str,
			const std::map<std::string,std::string> & bindings)
{
    return exec(str,bindings,0,0);
}

bool Database::exec(	const std::string & str,
			const std::map<std::string,std::string> & bindings,
			void (*data)(void *, std::vector<std::string*> & ),void * bindarg)
{
    int rc;
    int rowcount;
    int bindCount;
    const char * zErrMsg;
    sqlite3_stmt *pStmt;
    char * zSql = (char*)str.c_str();
    
    do {
      /* Compile the statement from SQL. Assume success. */
      rc = sqlite3_prepare(ppDb, zSql, -1, &pStmt, &zErrMsg);
      if (rc != SQLITE_OK) {
	  jsrlog("%s: SQL error: %s\n", zSql, zErrMsg);
	  return false;
      }
      
      bindCount = sqlite3_bind_parameter_count(pStmt);
      for(int i = 1 ; i <= bindCount; ++i) {
	  
	  std::string name(sqlite3_bind_parameter_name(pStmt,i));
	  typeof(bindings.find(name)) it = bindings.find(name);
	  if (it == bindings.end()) {
	      jsrlog("%s: Undefined parameter %s\n",zSql,name.c_str());
	      sqlite3_finalize(pStmt);
	      return false;
	  }
	  if (name.substr(0,3) == "$I_") {
	      long long v;
	      sscanf(it->second.c_str(),"%Ld",&v);
	      sqlite3_bind_int64(pStmt, i, v);
	  } else {
	      sqlite3_bind_blob(pStmt, i, it->second.data(), it->second.size(), SQLITE_TRANSIENT);
	  }
      }
      
      rowcount = sqlite3_column_count(pStmt);
      
      while( SQLITE_ROW == sqlite3_step(pStmt) ){
	  if (data){
	      std::vector<std::string *> values;
	      /* Do something with the row of available data */
	      for(int i = 0; i < rowcount; ++i) {
		  int sze = sqlite3_column_bytes(pStmt,i);
		  std::string * str = new std::string((char*)sqlite3_column_blob(pStmt,i),sze);
		  values.push_back(str);
	      }
	      (*data)(bindarg,values);
	      for(int i = 0; i < rowcount; ++i) {
		  if (values[i] != 0) {
		      delete(values[i]);
		  }
	      }
	  }
      }

      /* Finalize the statement. If an SQLITE_SCHEMA error has
      ** occured, then the above call to sqlite3_step() will have
      ** returned SQLITE_ERROR. sqlite3_finalize() will return
      ** SQLITE_SCHEMA. In this case the loop will execute again.
      */
      rc = sqlite3_finalize(pStmt);
    } while( rc == SQLITE_SCHEMA );
    return true;
}

void Database::initDatabase()
{
    dbTaskQueue = InternalTask::newAsyncTaskQueue(1);
}

TaskQueue * Database::dbTaskQueue = 0;
