#ifndef GRAPHS_H

#define GRAPHS_H

#include <string>

#include <SDL_ttf.h>

#include "jsreflect.h"

void SDL_stretch(SDL_Surface * org,SDL_Surface * dest,int ptcoords[4][2]);
void SDL_ComposeSurface(SDL_Surface *src, SDL_Rect *vsrcrect, SDL_Surface *dst, SDL_Rect *dstrect);
void SDL_ComposeBox(SDL_Surface *dst, SDL_Rect *dstrect,Uint32 color);
void SDL_AdjustAlpha(SDL_Surface *dst, SDL_Rect *dstrect, int level);

Uint32 getpixel(SDL_Surface *surface, int x, int y);
void putpixel(SDL_Surface *surface, int x, int y, Uint32 pixel);

class Surface;
class Painter;
class Widget;

typedef short int ScreenCoord;

typedef struct Rectangle
{
    ScreenCoord X, Y;
    ScreenCoord W, H;
    Rectangle() {};
    Rectangle(int x,int y,int w,int h) { X = x; Y = y; W = w; H = h; };
};

class ScreenArea
{
private:
    void AddSubstracting( Rectangle r, const ScreenArea * substracter, int substractpos );
protected:

    int NewRectangle();
    void DelRectangle( int id );

    bool GetRectangleIntersection( const Rectangle * src, Rectangle * Result ) const;
public:
    Rectangle* Rectangles;
    int RectangleCount;

    ScreenArea();
    ScreenArea( int x, int y, int w, int h );
    ScreenArea( const ScreenArea & cpy );
    ~ScreenArea();
    bool IsEmpty() const;
    ScreenArea* Intersect( const ScreenArea * copy ) const;
    ScreenArea* Add( const ScreenArea * copy ) const;
    
    ScreenArea* Substract( const ScreenArea * copy ) const;
    
    void Add(int x0,int y0,int w,int h);
    
    bool Contact( const Rectangle & r ) const;
    bool Contact( const ScreenArea & r ) const;
    
    void bounding(int & minx,int & miny,int & w,int & h) const;
	
    void Debug() const;
};

class ScreenArea;

class Color {
public:
    Uint32 color;
    
    Color(int r = 0,int g = 0,int b = 0,int a = 255);
    
    void setR(int r);
    int r() const { return (color >> 24) & 255; }
    void setG(int g);
    int g() const { return (color >> 16) & 255; }
    void setB(int b);
    int b() const { return (color >> 8) & 255; }
    void setA(int a);
    int a() const { return color & 255; }
    
    SDL_Color as_SDL_Color() const
    {
	SDL_Color ret = {r(),g(),b(),a()};
	return ret;
    }
    
    Uint32 as_uint() const
    {
	return color;
    }
    
    Uint32 map(SDL_Surface * s) const;
    
    static Color mix(const Color & a,const Color & b,int cur,int max);
};

/// dont use color key for frequently modified surfaces.

class Surface
{
    friend class Painter;
public:
    typedef enum FormatFlag { Opaque, ColorKey, Alpha };
protected:
    FormatFlag Format;
    bool DisplayTuned;
    
    SDL_Surface * Surf;
    
public:
    Surface();
    virtual ~Surface();
    
    void encode(std::string & to);
    void decode(const std::string & from);
    
    /// Check wether the given pixel is completly transparent.
    bool IsClear(int x,int y) const;
    
    bool IsPaletted() const;
    
    int MapRGB(int r,int g,int b) const;
    
    Uint32 mapColor(const Color & c) const;
    
    void clear(Uint32 c);
    
    /// Check wether the given surface has transparency.
    bool HasTransparency() const;
    
    /// Width of the surface
    int GetWidth() const;
    
    /// Height of the surface
    int GetHeight() const;
    
    /// Create a surface
    void Create(int width,int height, FormatFlag format, bool displayTune = false, int ckey = -1);
    
    /// Use for primary (screen) surface.
    void assign(SDL_Surface * s);
    
    /// Load an image, without transparency
    bool LoadImage(const char * name, FormatFlag format, bool displayTune = true, int ckey = -1);
    
    /// Set the colorkey (unused)
    void SetColorKey(int color);
    
    /// Set the global surface transparency (work only for Opaque Surfaces)
    void SetTransparency(int color);
    
    /// Return the global surface transparency
    int getTranparency() const;
    
    void DisplayTune();
    
    SDL_Surface * GetSDLSurface() const;
    
protected:
    void LooseSurface();
    void Convert(FormatFlag format, bool displayTuned, bool displayTune);
};


class Font : public JSReflection::Object
{
    JS_EXPORTED_CLASS
    
    TTF_Font * TTFFont;
    static bool initialised;
public:
    Font();
    virtual ~Font();
    
    bool Load(const char * filename,int size);
    
    void SetStyle(int style);
    int GetStyle() const;
    
    bool IsBold() const;
    void SetBold(bool b);
    
    bool IsItalic() const;
    void SetItalic(bool b);
    
    bool IsUnderline() const;
    void SetUnderline(bool b);

    const char * GetName();
    int GetSize() const;
    
    TTF_Font * GetFont() const;
    
    /// Compute the size of a text.
    void sizeText(const std::string & text, int & w, int & h) const;
    
    static Font * LoadFont(const char * filename,int size);
};

class Painter
{
protected:
    bool OwnSurf;
    SDL_Surface * Surf;
    
    bool HasArea;
    bool ownArea;
    const ScreenArea * Area;

    int XShift, YShift;
    int XSurfShift,YSurfShift;
    bool Intersect( const Rectangle & relativeTarget,
		    const Rectangle & target, 
		    SDL_Rect & realTarget, SDL_Rect & resulting);
    
    void SDLBlit(SDL_Surface * f,int x0,int y0,int w=-1,int h=-1,int surfx0=0,int surfy0=0);
public:
    Painter(Surface * surface,const ScreenArea * dc);
    Painter(Widget * w,Surface * dest,const ScreenArea * dc);
    virtual ~Painter();
    
    /// Set the position of the (0,0) point in the surface, on screen
    void SetSurfaceRelative(int x,int y);
    void SetSurfaceRelativeX(int x);
    void SetSurfaceRelativeY(int x);
    int GetSurfaceRelativeX() const;
    int GetSurfaceRelativeY() const;
    
    /// Set the position in the surface for the (0,0) point
    void SetRelative(int x,int y);
    void SetRelativeX(int x);
    void SetRelativeY(int x);
    int GetRelativeX() const;
    int GetRelativeY() const;
    
    //int Color(int r,int g,int b);
    
    
    void AdjustAlpha(int x0,int y0,int w,int h,int alphalevel);
    void FillRectangle(int x0,int y0,int w,int h,const Color & c);
    void Blit(Surface * f,int x0,int y0,int w=-1,int h=-1,int surfx0=0,int surfy0=0);
    void DrawText(const char * text,int x,int y,Font * f,const Color & c,int halign,int valign);
    
    void shift(int x,int y);
    const ScreenArea * getArea() const;
};


#endif
