#include <SDL.h>
#include <SDL_image.h>

#include "Graphs.h"
#include "Application.h"
#include "ImageWidget.h"

#include <math.h>

#include <string>

ImageWidget::ImageWidget() : Widget()
{
    image.setTarget(this);
    centerX = 0;
    centerY = 0;
    angle = 0;
    zoom  = 256;
    currentSurface = 0;
}

ImageWidget::~ImageWidget()
{
    if (currentSurface) {
	delete(currentSurface);
    }
}

static double orientations[4][2]= { {-1,-1} , {-1,1}, {1,-1}, {1,1}};

ImageWidgetCoordinates ImageWidget::prepareDraw()
{
    ImageWidgetCoordinates result;
    result.destw = getWidth();
    result.desth = getHeight();
     
    Surface * backsurf = image.getSurface();
    result.surface = backsurf;
    result.surfSerial = image.getSerial();
    if (backsurf) {
	/*
	getWidth / 2 = v.( 0.5 . (imgsx - 1) . cos + 0.5 . (imgsy - 1) .sin)
	getHeight / 2 = v.( 0.5 . (imgsy - 1) . sin - 0.5 . (imgsx - 1) .cos)
	*/
	int imgsx,imgsy;
	double a,b,a1,a2,b1,b2;
	double v;
	result.exists = true;
	result.surfw = backsurf->GetWidth();
	result.surfh = backsurf->GetHeight();
	imgsx = backsurf->GetWidth();
	imgsy = backsurf->GetHeight();
	a1 = fabs(((imgsx - 1) * getCos() + (imgsy - 1) *getSin())) / 2;
	b1 = fabs(((imgsy - 1) * getCos() - (imgsx - 1) *getSin())) / 2;
	
	// Invert imgsy
	a2 = fabs(((imgsx - 1) * getCos() + (-(imgsy - 1) *getSin()))) / 2;
	b2 = fabs(((-(imgsy - 1) * getCos()) - (imgsx - 1) *getSin())) / 2;
	
	a = (a1 > a2 ? a1 : a2);
	b = (b1 > b2 ? b1 : b2);
	/*
	max v : v.a <= result.destw/2 and v.b <= h/2  
	 =>  v = w/(2*a) or v = h / (2*b)
	w/2*a > h / 2*b, ssi
	w * b / 2 > h * a / 2
	*/
	
	if (result.destw * b > result.desth * a) {
	    v = result.desth / (2 * b);
	} else {
	    v = result.destw / (2 * a);
	}
	v *= zoom / 256.0;
	/*
	Image projection is :
	[imgx,imgy] => [ptx,pty]
	ptx = v * ( (imgx - ctx - (imgsx - 1) / 2) * getCos + (imgy - cty - (imgsy - 1) /2) * getSin )
	pty = v * ( (imgy - cty - (imgsy - 1) / 2) * getCos - (imgx - ctx - (imgsx - 1) /2) * getSin )
	
	Next step is : find imgx,imgy value for ptx (-w/2,-h/2) (w/2,-h/2) (w/2,h/2) (-w/2,h/2)
	*/
	
	double ctx,cty;
	
	ctx = (centerX / 65536.0) * imgsx / 2.0;
	cty = (centerY / 65536.0) * imgsy / 2.0;
	
	for(int i = 0 ; i < 4 ; ++i) {
	    double ptx,pty;
	    double imgx,imgy;
	    
	    ptx = (result.destw * orientations[i][0]) / 2.0;
	    pty = (result.desth * orientations[i][1]) / 2.0;

	    imgx = (ctx + (((getCos() * ptx) - (getSin() * pty)) / (v * ((getSin() * getSin()) + (getCos() * getCos())))) + ((imgsx - 1.0) / 2.0));
	    imgy = (((imgsy - 1.0) / 2.0) + cty + (((getCos() * pty) + (getSin() * ptx)) / (v * ((getSin() * getSin()) + (getCos() * getCos())))));
	    result.ptcoords[i][0] = (int)(imgx * 256);
	    result.ptcoords[i][1] = (int)(imgy * 256);
	}
	// if zoom = 256, ptcoord[0][] must be (0,0), ptcoord[2][] must be (backsurf.w - 1) * 256, (backsurf.h - 1) * 256,  
    } else {
	result.exists = false;
    }
    return result;
}

/*
 * Return the pixel value at (x, y)
 * NOTE: The surface must be locked before calling this!
 */
Uint32 getpixel(SDL_Surface *surface, int x, int y)
{
    int bpp = surface->format->BytesPerPixel;
    /* Here p is the address to the pixel we want to retrieve */
    Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;
    
    switch(bpp) {
    case 1:
        return *p;

    case 2:
        return *(Uint16 *)p;

    case 3:
        if(SDL_BYTEORDER == SDL_BIG_ENDIAN)
            return p[0] << 16 | p[1] << 8 | p[2];
        else
            return p[0] | p[1] << 8 | p[2] << 16;

    case 4:
        return *(Uint32 *)p;

    default:
	fprintf(stderr,"getPixel : unknown depth : %d\n",bpp);
        return 0;       /* shouldn't happen, but avoids warnings */
    }
}

/*
 * Set the pixel at (x, y) to the given value
 * NOTE: The surface must be locked before calling this!
 */
void putpixel(SDL_Surface *surface, int x, int y, Uint32 pixel)
{
    int bpp = surface->format->BytesPerPixel;
    /* Here p is the address to the pixel we want to set */
    Uint8 *p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;

    switch(bpp) {
    case 1:
        *p = pixel;
        break;

    case 2:
        *(Uint16 *)p = pixel;
        break;

    case 3:
        if(SDL_BYTEORDER == SDL_BIG_ENDIAN) {
            p[0] = (pixel >> 16) & 0xff;
            p[1] = (pixel >> 8) & 0xff;
            p[2] = pixel & 0xff;
        } else {
            p[0] = pixel & 0xff;
            p[1] = (pixel >> 8) & 0xff;
            p[2] = (pixel >> 16) & 0xff;
        }
        break;

    case 4:
        *(Uint32 *)p = pixel;
        break;
    }
}

static int divmul(int v0,int v1,int cur,int max)
{
    return v0 + ((v1 - v0) * cur) / max;
}

void SDL_stretch(SDL_Surface * org, SDL_Surface * dest, int ptcoords[4][2])
{
    SDL_LockSurface(org);
    SDL_LockSurface(dest);
    for(int y = 0; y < dest->h; ++y) {
	int img_start[2],img_stop[2];
	img_start[0]=divmul(ptcoords[0][0],ptcoords[1][0],y,dest->h);
	img_start[1]=divmul(ptcoords[0][1],ptcoords[1][1],y,dest->h);
	img_stop[0]=divmul(ptcoords[2][0],ptcoords[3][0],y,dest->h);
	img_stop[1]=divmul(ptcoords[2][1],ptcoords[3][1],y,dest->h);
	for(int x = 0; x < dest->w; ++x) {
	    int img_cur[2];
	    Uint32 pixel;
	    img_cur[0] = divmul(img_start[0],img_stop[0],x,dest->w) / 256;
	    img_cur[1] = divmul(img_start[1],img_stop[1],x,dest->w) / 256;
	    
	    if ((img_cur[0] < 0) || (img_cur[1] < 0) || (img_cur[0] >= org->w) || (img_cur[1] >= org->h)) {
		pixel = 0;
	    } else {
		pixel = getpixel(org,img_cur[0],img_cur[1]);
	    }
	    
	    putpixel(dest,x,y,pixel);
	}
    }
    SDL_UnlockSurface(org);
    SDL_UnlockSurface(dest);
}


void ImageWidget::createSurface(SDL_Surface * org,SDL_Surface * dest)
{
    if (!lastdraw.exists) {
	return;
    }
    jsrlog("CreateSurface  from %dx%d to %dx%d: src is %d, dst is %d\n", dest->w, dest->h, org->w, org->h, org->format->BytesPerPixel, dest->format->BytesPerPixel);
    SDL_stretch(org,dest,lastdraw.ptcoords);
}

void ImageWidget::drawSelf(Painter & p)
{
    ImageWidgetCoordinates nvconfig;
    nvconfig = prepareDraw();
    
    if (!nvconfig.compare(lastdraw)) {
	// Recreate surface
	lastdraw = nvconfig;
	if (currentSurface) {
	    delete(currentSurface);
	    currentSurface = 0;
	}
	SDL_Surface * org = (image.getSurface() ? lastdraw.surface->GetSDLSurface() : 0);
	if (org) {
	    currentSurface = new Surface();
	    currentSurface->Create(getWidth(),getHeight(),Surface::Alpha);
	    createSurface(org,currentSurface->GetSDLSurface());
	}
    }
    if (currentSurface) {
	p.Blit(currentSurface,0,0);
    }
    
    Widget::drawSelf(p);
}

void ImageWidget::notified(Notifier * n)
{
    jsrlog("ImageWidget %#x notified by %#x\n",(Widget*)this,n);
    if (n == &image) {
	if (checkScreenUpdate() && isGlobalVisible()) {
	    askScreenUpdate();
	}
    } else {
	Widget::notified(n);
    }
}

static int exact_cos[4][2]= {{1,0},{0,1},{-1,0},{0,-1}};

double ImageWidget::getCos() const
{
    // 64 is clock-wise
    int agl = angle & 63;
    if ((agl & 15) == 0) {
	agl /= 16;
	return exact_cos[agl][0];
    }
    return cos(angle*2*M_PI / 64);
}

double ImageWidget::getSin() const
{
    int agl = angle & 63;
    if ((agl & 15) == 0) {
	agl /= 16;
	return exact_cos[agl][1];
    }
    return sin(angle*2*M_PI / 64);
}

int ImageWidget::getCenterX() const 
{ return centerX; }

void ImageWidget::setCenterX(int u)
{ 
    centerX = u; 
    if (checkScreenUpdate() && isGlobalVisible()) {
	askScreenUpdate();
    }    
}

int ImageWidget::getCenterY() const 
{ return centerY; }

void ImageWidget::setCenterY(int u)
{ 
    centerY = u;
    if (checkScreenUpdate() && isGlobalVisible()) {
	askScreenUpdate();
    }
}

int ImageWidget::getAngle() const 
{ return angle; }

void ImageWidget::setAngle(int u)
{ 
    if (checkScreenUpdate() && isGlobalVisible()) {
	askScreenUpdate();
    }
    angle = u; 
}

int ImageWidget::getZoom() const 
{ return zoom; }

void ImageWidget::setZoom(int u)
{ 
    zoom = u;
    if (checkScreenUpdate() && isGlobalVisible()) {
	askScreenUpdate();
    }
}

int ImageWidget::getVectorXX() 
{
    ImageWidgetCoordinates coords = prepareDraw();
    if (!coords.exists) {
	return 0;
    }
    return (coords.ptcoords[2][0] - coords.ptcoords[0][0]) *256.0 / coords.surfw;
}

int ImageWidget::getVectorXY() 
{
    ImageWidgetCoordinates coords = prepareDraw();
    if (!coords.exists) {
	return 0;
    }
    return (coords.ptcoords[2][1] - coords.ptcoords[0][1]) *256.0 / coords.surfh;
}

int ImageWidget::getVectorYX() 
{
    ImageWidgetCoordinates coords = prepareDraw();
    if (!coords.exists) {
	return 0;
    }
    return (coords.ptcoords[1][0] - coords.ptcoords[0][0]) *256.0 / coords.surfw;
}

int ImageWidget::getVectorYY() 
{
    ImageWidgetCoordinates coords = prepareDraw();
    if (!coords.exists) {
	return 0;
    }
    return (coords.ptcoords[1][1] - coords.ptcoords[0][1]) *256.0 / coords.surfh;
}

void ImageWidget::globalVisibleChanged()
{
    Widget::globalVisibleChanged();
    
    if (isGlobalVisible()) {
	if (image_file == "") {
	    image.bind(0);
	} else {
	    ImageFactory::loadImage(image_file,image);
	}
	
	// FIXME : update when image is loaded as well !!!
	if (checkScreenUpdate() && isGlobalVisible()) {
	    askScreenUpdate();
	}
    } else {
	image.bind(0);
    }

}

void ImageWidget::setImage(const std::string & s) {
    image_file = s;
    if (isGlobalVisible()) {
	if (s == "") {
	    image.bind(0);
	} else {
	    ImageFactory::loadImage(s,image);
	}

	// FIXME : update when image is loaded as well !!!
	if (checkScreenUpdate() && isGlobalVisible()) {
	    askScreenUpdate();
	}
    } else {
	image.bind(0);
    }
}

std::string ImageWidget::getImage() const {
    if (image.binded()) {
	return image.getProducer()->getName();
    } 
    return "";
}



ImageWidgetCoordinates::ImageWidgetCoordinates()
{
    exists = false;
    surfSerial = 0;
}

JS_EXPORT_CLASS(ImageWidget,Widget);
JS_RW_PROPERTY(ImageWidget,angle,getAngle,setAngle,int,"%i",0);
JS_RW_PROPERTY(ImageWidget,zoom,getZoom,setZoom,int,"%i",0);
JS_RW_PROPERTY(ImageWidget,centerX,getCenterX,setCenterX,int,"%i",0);
JS_RW_PROPERTY(ImageWidget,centerY,getCenterY,setCenterY,int,"%i",0);
JS_RW_PROPERTY(ImageWidget,image,getImage,setImage,JSReflection::JS_String,"%s",0);
JS_RO_PROPERTY(ImageWidget,vectorXX,getVectorXX,int,"%i",0);
JS_RO_PROPERTY(ImageWidget,vectorXY,getVectorXY,int,"%i",0);
JS_RO_PROPERTY(ImageWidget,vectorYX,getVectorYX,int,"%i",0);
JS_RO_PROPERTY(ImageWidget,vectorYY,getVectorYY,int,"%i",0);

