#include "Application.h"
#include "Widget.h"
#include "Layout.h"

Layout::Layout() : JSReflection::Object(Application::getJsContext())
{
    widget = 0;
}

Layout::~Layout()
{ }

Widget * Layout::getArea() const
{
    return widget;
}

FullLayout::FullLayout() : Layout()
{
    vmargin = 0;
    hmargin = 0;
}

FullLayout::~FullLayout()
{}

void FullLayout::addWidget(Widget * v)
{
    v->layedOut = true;
    relayoutWidget(v);
}

void FullLayout::relayoutWidget(Widget * v)
{
    if (!v->layedOut) return;
    v->setPosition(hmargin,vmargin,widget->getWidth() - 2*hmargin,widget->getHeight()-2*vmargin);
}

void FullLayout::update()
{}

GridLayout::GridLayout() : Layout()
{
    curline = 0;
    curcell = 0;
}

GridLayout::~GridLayout()
{
    for(std::vector<LayoutLine*>::const_iterator it = lines.begin();it != lines.end(); ++it) {
	delete(*it);
    }
}
void GridLayout::update()
{
    for(int curline = 0; curline < (signed)lines.size(); ++curline){
	lines[curline]->update();
    }
}

void GridLayout::relayoutWidget(Widget * widget)
{
    int cline,ccell;
    int x,y,w,h;
    LayoutLine * line;
    
    if (!widget->layedOut) { return; }
    
    ccell = widget->layoutId1;
    cline = widget->layoutId2;
    
    y = 0;
    for(int i = 0; i < cline; ++i) {
	y += lines[i]->getHeight();
    }
    line = lines[cline];
    y += line->getTopBorder();
    h = line->getHeight() - line->getTopBorder() - line->getBottomBorder();
    if (h < 0) { h = 0; }
    
    x = 0;
    for(int i = 0 ; i < ccell; ++i) {
	x += line->getCell(i)->getWidth();
    }
    LayoutCell * cell = line->getCell(ccell);
    x += cell->getLeftBorder();
    w = cell->getWidth() - cell->getLeftBorder() - cell->getRightBorder();
    
    if (w < 0) { w = 0; }
    
    widget->setPosition(x,y,w,h);
}

void GridLayout::addWidget(Widget * widget)
{
    
    if (curline >= (signed)lines.size()) return;
    widget->layedOut = true;
    widget->layoutId1 = curcell;
    widget->layoutId2 = curline;
    relayoutWidget(widget);
    
    // Next line, svp
    curcell++;
    if (curcell < lines[curline]->getCellCount()) {
	return;
    }
    curcell = 0;
    curline++;
    while ((curline < (signed)lines.size()) && (!lines[curline]->getCellCount())) {
	curline++;
    }
}

void GridLayout::addLine(LayoutLine * l)
{
    lines.push_back(l);
    l->layout = this;
}

LayoutLine::LayoutLine() : JSReflection::Object(Application::getJsContext())
{
    height=100;
    top_border = 0;
    bottom_border = 0;
    layout = 0;
}

LayoutLine::~LayoutLine()
{
    for(std::vector<LayoutCell*>::const_iterator it = cells.begin();it != cells.end(); ++it) {
	delete(*it);
    }
}

Widget * LayoutLine::getArea() const
{
    if (layout) {
	return layout->getArea();
    }
    return 0;
}

void LayoutLine::addCell(LayoutCell * cell)
{
    cells.push_back(cell);
    cell->layout = layout;
}

int LayoutLine::getCellCount() const
{
    return cells.size();
}

LayoutCell * LayoutLine::getCell(int i) const
{
    return cells[i];
}

int LayoutLine::getHeight() const 
{ return height; }

void LayoutLine::setHeight(int u)
{ height = u; }

int LayoutLine::getTopBorder() const 
{ return top_border; }

void LayoutLine::setTopBorder(int u)
{ top_border = u; }

int LayoutLine::getBottomBorder() const 
{ return bottom_border; }

void LayoutLine::setBottomBorder(int u)
{ bottom_border = u; }


void LayoutLine::setStrHeight(const std::string & u)
{
    str_height = u;
}

void LayoutLine::setStrTopBorder(const std::string & u)
{
    str_top_border = u;
}

void LayoutLine::setStrBottomBorder(const std::string & u)
{
    str_bottom_border = u;
}

LayoutCell::LayoutCell() : JSReflection::Object(Application::getJsContext())
{
    width = 10;
    left_border = 0;
    right_border = 0;
}

LayoutCell::~LayoutCell()
{}

Widget * LayoutCell::getArea() const
{
    if (layout) {
	return layout->getArea();
    }
    return 0;
}

int LayoutCell::getWidth() const 
{ return width; }

void LayoutCell::setWidth(int u)
{ width = u; }

int LayoutCell::getLeftBorder() const 
{ return left_border; }

void LayoutCell::setLeftBorder(int u)
{ left_border = u; }

int LayoutCell::getRightBorder() const 
{ return right_border; }

void LayoutCell::setRightBorder(int u)
{ right_border = u; }


void LayoutCell::setStrWidth(const std::string & u)
{
    str_width = u;
}

void LayoutCell::setStrLeftBorder(const std::string & u)
{
    str_left_border = u;
}

void LayoutCell::setStrRightBorder(const std::string & u)
{
    str_right_border = u;
}

void LayoutCell::update()
{
    if (str_width.length()) setProperty("width",str_width.c_str(),definition);
    if (str_left_border.length()) setProperty("leftBorder",str_left_border.c_str(),definition);
    if (str_right_border.length()) setProperty("rightBorder",str_right_border.c_str(),definition);
}

void LayoutLine::update()
{
    if (str_height.length()) setProperty("height",str_height.c_str(),definition);
    if (str_top_border.length()) setProperty("topBorder",str_top_border.c_str(),definition);
    if (str_bottom_border.length()) setProperty("bottomBorder",str_bottom_border.c_str(),definition);
    for(int i = 0; i < (signed)cells.size(); ++i) {
	cells[i]->update();
    }
}


JS_EXPORT_CLASS(Layout,JSReflection::Object)
JS_RO_PROPERTY(Layout,area,getArea,Widget*,"%x",Widget::ClassInfo);

JS_EXPORT_CLASS(LayoutLine,JSReflection::Object)
JS_RO_PROPERTY(LayoutLine,area,getArea,Widget*,"%x",Widget::ClassInfo);
JS_RW_PROPERTY(LayoutLine,height,getHeight,setHeight,int,"%i",0);
JS_RW_PROPERTY(LayoutLine,topBorder,getTopBorder,setTopBorder,int,"%i",0);
JS_RW_PROPERTY(LayoutLine,bottomBorder,getBottomBorder,setBottomBorder,int,"%i",0);

JS_EXPORT_CLASS(LayoutCell,JSReflection::Object)
JS_RO_PROPERTY(LayoutCell,area,getArea,Widget*,"%x",Widget::ClassInfo);
JS_RW_PROPERTY(LayoutCell,width,getWidth,setWidth,int,"%i",0);
JS_RW_PROPERTY(LayoutCell,leftBorder,getLeftBorder,setLeftBorder,int,"%i",0);
JS_RW_PROPERTY(LayoutCell,rightBorder,getRightBorder,setRightBorder,int,"%i",0);

JS_EXPORT_CLASS(FullLayout,Layout)
JS_RW_PROPERTY(FullLayout,hmargin,getHMargin,setHMargin,int,"%i",0);
JS_RW_PROPERTY(FullLayout,vmargin,getVMargin,setVMargin,int,"%i",0);


