#include <SDL.h>
#include <SDL_thread.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <stdexcept>

#include "jsreflect.h"

#include "Metadata.h"
#include "Database.h"

// We use the dbTaskQueue, which garanty that two run will not 
// be executed consecutively.

std::map<std::string,MetaProvider *> MetaUser::metadata;
static Database * metaDb = 0;


MetaProvider::MetaProvider(const std::string & p):InternalTask(Database::dbTaskQueue),path(p)
{
    jsrlog("MetaProvider created for %s\n",p.c_str());
    usageCount = 0;
    ready = false;
    needUpdate = false;
}

MetaProvider::~MetaProvider() 
{
    // FIXME : 
    jsrlog("MetaProvider deleted for %s\n",path.c_str());
    MetaUser::metadata.erase(path);
}

bool MetaProvider::isReady()
{
    bool r;
    InternalTask::lock();
    r = ready;
    InternalTask::unlock();
    return r;
}

void MetaProvider::dump(void *d, std::vector<std::string*> & values)
{
    InternalTask::lock();
    
    MetaProvider * mp = (MetaProvider*)d;
    if (mp->cancelAction) return;
    jsrlog("%s got key %s\n",mp->path.c_str(),(*values[0]).c_str());
    mp->values[(*values[0])] = (*values[1]);
    InternalTask::unlock();
}

static void setstring(void *a, std::vector<std::string*> & v)
{
    std::string * str = (std::string *)a;
    (*str) = (*v[0]);
}

bool MetaProvider::suicide()
{
    return (!usageCount)&(!isRunning())&&(!pendingSignal());
}

void MetaProvider::fileChanged()
{
    InternalTask::lock();
    if (isRunning()) {
	cancelAction = true;
	if (!ready) {
	    emitNotification();
	}
	ready = true;
	autoSignal();
    }
    values.clear();
    InternalTask::unlock();
}

static std::string getFileTag(const std::string & encodedPath)
{
    struct stat buf;
    char buffer[512];
    std::string path = JSReflection::decodeBinaryPath(encodedPath);
    
    if (stat(path.c_str(), &buf) == -1) {
	return "";
    }
    snprintf(buffer,512,"%llu-%llu",(long long unsigned)buf.st_mtime,(long long unsigned)buf.st_size);
    return std::string(buffer);
}

void MetaProvider::run(TaskNotification * tn)
{
    cancelAction = false;
    
    if (!metaDb) {
	jsrlog("Creating new metadata database\n");
	InternalTask::unlock();
	metaDb = new Database("plop.db");
	if (!metaDb->open()) {
	    jsrlog("Creating new metadata database failed !\n");
	    delete(metaDb);
	    InternalTask::lock();
	    metaDb = 0;
	    ready = true;
	    emitNotification();
	    return;
	}
	// Initialise...
	metaDb->exec("CREATE TABLE metas (id INTEGER PRIMARY KEY AUTOINCREMENT, name TEXT NOT NULL UNIQUE,tag TEXT NOT NULL);");
	metaDb->exec("CREATE TABLE metavalues (id INTEGER NOT NULL, name TEXT NOT NULL, value BLOB NOT NULL, PRIMARY KEY(id,name));");
	metaDb->exec("CREATE INDEX metaindex ON metavalues (id,name);");
	InternalTask::lock();
	jsrlog("New metadata database created\n");
    }
    
    if (!ready) {
	jsrlog("%s MetaProvider loading...\n",path.c_str());
	
	// If db not opened, open it
	InternalTask::unlock();
	
	std::string fileTag = getFileTag(path);
	
	std::map<std::string,std::string> args;
	args["$S_NAME"] = path;
	
	if (fileTag != "") {
	    args["$S_TAG"] = fileTag;
	
	    metaDb->exec("DELETE FROM metavalues WHERE id in (SELECT id FROM metas WHERE metas.name=$S_NAME AND metas.tag!=$S_TAG)",args);
	    metaDb->exec("DELETE FROM metas WHERE metas.name=$S_NAME and metas.tag!=$S_TAG",args);
	    metaDb->exec("SELECT metavalues.name,metavalues.value FROM metas,metavalues WHERE metas.id=metavalues.id and metas.name=$S_NAME",args,&MetaProvider::dump,this);
	} else {
	    jsrlog("loading metadata for non existent file : %s\n",path.c_str());
	    metaDb->exec("DELETE FROM metavalues WHERE id in (SELECT id FROM metas WHERE metas.name=$S_NAME)",args);
	    metaDb->exec("DELETE FROM metas WHERE metas.name=$S_NAME",args);
	}
	InternalTask::lock();
	if (!cancelAction) {
	    jsrlog("MetaProvider loaded...\n");
	    ready = true;
	    emitNotification();
	}
    } else {
	std::string id;
	std::map<std::string,std::string> args;
	jsrlog("MetaProvider saving...\n");
	
	InternalTask::unlock();
	std::string fileTag = getFileTag(path);
	
	args["$S_NAME"] = path;
	args["$S_TAG"] = fileTag;
	metaDb->exec("DELETE FROM metavalues WHERE id in (SELECT id FROM metas WHERE metas.name=$S_NAME)",args);
	metaDb->exec("DELETE FROM metas WHERE metas.name=$S_NAME",args);
	if (fileTag != "") {
	    metaDb->exec("INSERT INTO metas (name,tag) VALUES ($S_NAME,$S_TAG)",args);
	    metaDb->exec("SELECT id FROM metas WHERE name=$S_NAME and tag=$S_TAG",args,setstring,&id);

	    InternalTask::lock();
	    if (cancelAction) {
		return;
	    }
	    std::map<std::string,std::string> tmpvalues = values;
	    InternalTask::unlock();
	    args["$I_ID"] = id;
	    
	    for(typeof(tmpvalues.begin()) it = tmpvalues.begin(); it != tmpvalues.end(); ++it) {
		std::map<std::string,std::string> args;
		
		InternalTask::lock();
		if (cancelAction) {
		    return;
		}
		InternalTask::unlock();
		
		args["$S_NAME"] = it->first;
		args["$S_VALUE"] = it->second;
		args["$I_ID"] = id;
		jsrlog("(ID %s) Saving key %s, size=%d\n",id.c_str(),it->first.c_str(),args["$S_VALUE"].size());
		metaDb->exec("INSERT INTO metavalues (id,name,value) VALUES ($I_ID,$S_NAME,$S_VALUE)",args);
	    }
	}
	InternalTask::lock();
    }
}

bool MetaProvider::getKey(const std::string & name, std::string & val)
{
    InternalTask::lock();
    if (!ready) {
	InternalTask::unlock();
	return false;
    }
    if (values.find(name) != values.end()) {
	val = values[name];
	jsrlog("%s: key %s got [%d bytes]\n",path.c_str(),name.c_str(),val.size());
	InternalTask::unlock();
	return true;
    } else {
	InternalTask::unlock();
	return false;
    }
}

void MetaProvider::setKey(const std::string & name,const std::string & val)
{
    InternalTask::lock();
    if (ready) {
	jsrlog("key %s set to [%d bytes]\n",name.c_str(),val.size());
	values[name] = val;
	autoSignal();
    }
    InternalTask::unlock();
}

MetaUser::MetaUser(TaskQueue * which) : Notifier(which) 
{
    mp = 0;
}

MetaUser::~MetaUser()
{
    bind(0);
}

void MetaUser::bind(MetaProvider * m)
{
    if (m == mp) {
	return;
    }
    InternalTask::lock();
    if (mp) {
	mp->usageCount--;
	unwaitNotification(mp);
	if (mp->suicide()) {
	    delete(mp);
	}
    }
    if (m) {
	m->usageCount++;
	waitNotification(m);
    }
    mp = m;
    InternalTask::unlock();
}

void MetaUser::bind(const std::string path)
{
    MetaProvider * m;
    InternalTask::lock();
    jsrlog("Searching %s\n",path.c_str());
    if (metadata.find(path) == metadata.end()) {
	m = new MetaProvider(path);
	metadata[path] = m;
	m->autoSignal();
    } else {
	jsrlog("cache hit for %s\n",path.c_str());
	m = metadata[path];
    }
    bind(m);
    InternalTask::unlock();
}

bool MetaUser::binded()
{
    return mp != 0;
}

void MetaUser::unbind()
{
    bind(0);
}

bool MetaUser::isReady()
{
    if (mp) {
	return mp->isReady();
    } else {
	return 0;
    }
}

bool MetaUser::getKey(const std::string & name, std::string & val)
{
    if (mp) {
	return mp->getKey(name,val);
    } else {
	return false;
    }
    
}

void MetaUser::setKey(const std::string & name, const std::string & val)
{
    if (mp) {
	mp->setKey(name,val);
    }
}

void MetaUser::invalidateFile(const std::string & path)
{
    InternalTask::lock();
    typeof(metadata.find(path)) it = metadata.find(path);
    if (it != metadata.end()) {
	it->second->fileChanged();
	jsrlog("Signaled existing metadata.\n");
    } else {
	jsrlog("Created new empty metadata.\n");
	MetaProvider * mp = new MetaProvider(path);
	metadata[path] = mp;
	mp->ready = true;
	mp->needUpdate = true;
	mp->autoSignal();
    }
    InternalTask::unlock();
}
