#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include <SDL.h>
#include <SDL_image.h>

#include "Graphs.h"
#include "Application.h"
#include "Producer.h"
#include "Processes.h"
#include "XmlParser.h"
#include <string>

Selectable::Selectable()
{
    fd = -1;
}

Selectable::~Selectable()
{
    setFd(-1);
}

void Selectable::setFd(int nvfd)
{
    int oldfd = fd;
    fd = nvfd;
    
    if ((oldfd == -1) && (nvfd == -1)) {
	return;
    }
    
    if ((oldfd != -1) && (nvfd != -1)) {
	// change fd => Nothing todo
	return;
    }
    
    // Release fd
    if ((oldfd != -1)&&(nvfd == -1)) {
	Application::current->selectables.erase(this);
	return;
    }
    
    // Take a fd
    Application::current->selectables.insert(this);
}

int Selectable::getFd() const
{
    return fd;
}

Producer::Producer(const char * c, JSReflection::Object * vroot, JSReflection::JSClosure * ne, JSReflection::JSClosure * de,JSReflection::JSClosure * ce):
	Selectable(),JSReflection::Object(Application::getJsContext()),cmd(c)
{
    root = vroot;
    newEntry = ne;
    done = de;
    childExited = ce;
    errStatus = 0;
    isXmlConsummer = false;
    parser = 0;
    child=0;
}

void Producer::setXml()
{
    isXmlConsummer = true;
}

Producer::~Producer()
{
    if (parser) {
    	delete(parser);
    }
    if (getFd() != -1) {
	close(getFd());
	setFd(-1);
    }
    if (newEntry) {
	delete(newEntry);
    }
    if (done) {
	delete(done);
    }
    if (childExited) {
	delete(childExited);
    }
    if (child) {
	delete(child);
    }
}

void Producer::doRead()
{
    char tmp[4096];
    int ret;
    do {
	ret = read(getFd(),tmp,4096);
    } while ((ret == -1) && (errno == EINTR));
    
    if ((ret == 0) || (ret == -1)) {
	if (ret == -1) {
	    if (errno == EAGAIN) {
		return;
	    }
	    perror(cmd.c_str());
	}
	close(getFd());
	setFd(-1);
	
	// Terminated.
	if ((!isXmlConsummer) && (received.size() > 0)) {
	    sendData(received);
	}
	jsval t = INT_TO_JSVAL(0);
	if (done) {
	    done->Call(root,t);
	}
	return;
    }
     
    received += std::string(tmp,ret);
    parseReceived();
}

void Producer::sendData(const std::string & data)
{
    if (newEntry != 0) {
	JSString * str = 0;
	unsigned sze=data.size();
	if (sze) {
	    char * t = (char*)JS_malloc(Application::getJsContext()->GetJSContext(),sze);
	    memcpy(t,data.data(),sze);
	    str = JS_NewString(Application::getJsContext()->GetJSContext(),t,sze);
	} else {
	    str = JS_NewString(Application::getJsContext()->GetJSContext(),0,0);
	}
	newEntry->Call(root,STRING_TO_JSVAL(str));
    }
}

void Producer::parseReceived()
{
    if (isXmlConsummer) {
    	std::list<JSReflection::Object*>newentries;
        parser->parse(received.data(),received.length(),newentries);
        for(std::list<JSReflection::Object*>::const_iterator it=newentries.begin(); it!=newentries.end(); ++it) {
	    if (newEntry) {
		jsrlog("call newEntry %#x with it=%#x\n",newEntry,(JSReflection::Object*)(*it));
		newEntry->Call(root,*it);
	    }
	    // FIXME : *it may be deleted twice (?)
	    delete(*it);
        }
        received=std::string();
    } else {
	unsigned pos;
	while((pos = received.find('\n')) != std::string::npos) {
	    std::string tmp = received.substr(0, pos);
	    received = received.substr(pos+1);
	    sendData(tmp);
	}
    }
}

void Producer::notified(Notifier * n)
{
    if (n == child && childExited) {
	InternalTask::lock();
	jsval t = INT_TO_JSVAL(child->getReturnCode());
	child->detach();
	//delete(child);
	child = 0;
	InternalTask::unlock();
	childExited->Call(root,t);
    }
}

void Producer::cancel()
{
    // FIXME : kill process!
    if (getFd() != -1) {
	close(getFd());
	setFd(-1);
    }
    delete(child);
    child = 0;
}

void Producer::start()
{
    int filedes[2];
    
    cancel();
    
    if (isXmlConsummer) {
    	parser=new XmlParser();
	parser->start(cmd,Application::getJsContext());
    }
    
    if (pipe(filedes) == -1) {
	perror("pipe");
	errStatus = 1;
	return;
    }
    
    if (child) {
	delete(child);
    }
    child = new ChildProcess();
    child->setTarget(this);
    int t = child->fork();
    if (t == -1) {
	close(filedes[0]);
	close(filedes[1]);
	errStatus = 1;
	delete(child);
	child = 0;
	return;
    }
    if (t == 0) {
	close(filedes[0]);
	close(1);
	
	dup2(filedes[1],1);
	close(filedes[1]);
	
	execl("/bin/sh","/bin/sh","-c",JSReflection::decodeBinaryPath(cmd).c_str(),0);
	perror(cmd.c_str());
	exit(1);
    }
    close(filedes[1]);

    fcntl(filedes[0],F_SETFL,O_NONBLOCK);
    setFd(filedes[0]);
}

Producer * newProducer(const char * cmd,JSReflection::Object * obj,JSReflection::JSClosure * newe,JSReflection::JSClosure * done,JSReflection::JSClosure * terminated)
{
    Producer * p = new Producer(cmd,obj,newe,done,terminated);
    p->start();
    return p;
}

Producer * newXmlProducer(const char * cmd,JSReflection::Object * obj,JSReflection::JSClosure * newe,JSReflection::JSClosure * done)
{
    Producer * p = new Producer(cmd,obj,newe,done,0);
    p->setXml();
    p->start();
    return p;
}



JS_EXPORT_CLASS(Producer,JSReflection::Object)


JS_NOCONTEXT_GLOBAL_FUNCTION_5(newProducer,Producer*,"%x",Producer::ClassInfo,
				JSReflection::JS_String,"%s",0,
				JSReflection::Object*,"%x",JSReflection::Object::ClassInfo,
				JSReflection::JSClosure*,"%x",JSReflection::JSClosure::ClassInfo,
				JSReflection::JSClosure*,"%x",JSReflection::JSClosure::ClassInfo,
				JSReflection::JSClosure*,"%x",JSReflection::JSClosure::ClassInfo)
JS_NOCONTEXT_GLOBAL_FUNCTION_4(newXmlProducer,Producer*,"%x",Producer::ClassInfo,
                                JSReflection::JS_String,"%s",0,
                                JSReflection::Object*,"%x",JSReflection::Object::ClassInfo,
                                JSReflection::JSClosure*,"%x",JSReflection::JSClosure::ClassInfo,
				JSReflection::JSClosure*,"%x",JSReflection::JSClosure::ClassInfo)

JS_NOCONTEXT_OBJECT_VOID_FUNCTION_0(Producer,cancel);
