#include <SDL.h>
#include <SDL_image.h>

#include "Graphs.h"

#include <string>


#define RectangleStorageFor(a) ((a + 63) & (~63))

static bool RectangleIntersect( const Rectangle * a, const Rectangle * b, Rectangle * rslt )
{
    int a_min_x, a_min_y, a_max_x, a_max_y;
    int b_min_x, b_min_y, b_max_x, b_max_y;
    int x0, y0, x1, y1;
    
    a_min_x = a->X;
    a_min_y = a->Y;
    a_max_x = a->X + a->W - 1;
    a_max_y = a->Y + a->H - 1;
    
    b_min_x = b->X;
    b_min_y = b->Y;
    b_max_x = b->X + b->W - 1;
    b_max_y = b->Y + b->H - 1;
    
    x0 = ( a_min_x > b_min_x ? a_min_x : b_min_x );
    x1 = ( a_max_x < b_max_x ? a_max_x : b_max_x );
    if ( x1 < x0 )
    {
        return false;
    }
    y0 = ( a_min_y > b_min_y ? a_min_y : b_min_y );
    y1 = ( a_max_y < b_max_y ? a_max_y : b_max_y );
    if ( y1 < y0 )
    {
        return false;
    }

    rslt->X = x0;
    rslt->Y = y0;
    rslt->W = x1 - x0 + 1;
    rslt->H = y1 - y0 + 1;

    return true;
}

// Substract a Hole rectangle from src.
static int RectangleExplode( Rectangle & src, Rectangle & hole, Rectangle * rectangles )
{
    // We have :
    // ....    0221
    // .HH.    0HH1
    // .HH.    0HH1
    // ....    0331

    int rid;
    int x0_contact, y0_contact, x1_contact, y1_contact;

    x0_contact = ( src.X == hole.X );
    x1_contact = ( src.X + src.W - 1 == hole.X + hole.W - 1 );
    y0_contact = ( src.Y == hole.Y );
    y1_contact = ( src.Y + src.H - 1 == hole.Y + hole.H - 1 );

    rid = 0;
    // We need to create 4 rectangle
    if ( !x0_contact )
    {
        rectangles[ rid ].X = src.X;
        rectangles[ rid ].Y = src.Y;
        rectangles[ rid ].W = hole.X - src.X;
        rectangles[ rid ].H = src.H;
        rid++;
    }

    if ( !x1_contact )
    {
        rectangles[ rid ].X = hole.X + hole.W;
        rectangles[ rid ].Y = src.Y;
        rectangles[ rid ].W = src.X + src.W - ( hole.X + hole.W );
        rectangles[ rid ].H = src.H;
        rid ++;
    }

    if ( !y0_contact )
    {
        rectangles[ rid ].X = hole.X;
        rectangles[ rid ].Y = src.Y;
        rectangles[ rid ].W = hole.W;
        rectangles[ rid ].H = hole.Y - src.Y;
        rid++;
    }

    if ( !y1_contact )
    {
        rectangles[ rid ].X = hole.X;
        rectangles[ rid ].Y = hole.Y + hole.H;
        rectangles[ rid ].W = hole.W;
        rectangles[ rid ].H = src.Y + src.H - ( hole.Y + hole.H );
        rid++;
    }
    return rid;
}

void ScreenArea::AddSubstracting( Rectangle r, const ScreenArea * substracter, int substractpos )
{
    Rectangle intersection;

    for ( int i = substractpos; i < substracter->RectangleCount; i++ )
    {
        if ( RectangleIntersect( &r, substracter->Rectangles + i, &intersection ) )
        {

            Rectangle remaining[ 4 ];

            int remainingNb;

            remainingNb = RectangleExplode( r, intersection, remaining );

            // No result => totally hidden
            if ( remainingNb == 0 )
            {
                return ;
            }

            // Optionnaly recursively add when more than 1 result
            for ( int j = 1; j < remainingNb; j++ )
            {
                AddSubstracting( remaining[ j ], substracter, i + 1 );
            }

            // Just substitue & continue...
            r = remaining[ 0 ];
        }
    }
    // Simply add r.
    int rid = NewRectangle();
    Rectangles[ rid ] = r;
}

ScreenArea * ScreenArea::Intersect( const ScreenArea * copy ) const
{
    ScreenArea * rslt;
    int nextRect;

    if ( copy->RectangleCount > RectangleCount )
    {
        return copy->Intersect( this );
    }

    rslt = new ScreenArea();
    nextRect = rslt->NewRectangle();
    for ( int i = 0; i < RectangleCount;i++ )
    {
        for ( int j = 0; j < copy->RectangleCount; j++ )
        {
            if ( RectangleIntersect( Rectangles + i, copy->Rectangles + j, rslt->Rectangles + nextRect ) )
            {
                nextRect = rslt->NewRectangle();
            }
        }
    }

    rslt->DelRectangle( nextRect );
    return rslt;
}

ScreenArea * ScreenArea::Substract( const ScreenArea * substracter ) const
{
    ScreenArea * rslt;
    //Rectangle intersection;

    rslt = new ScreenArea();
    for ( int i = 0;i < RectangleCount; i++ )
    {
        rslt->AddSubstracting( Rectangles[ i ], substracter, 0 );
    }

    return rslt;
}

ScreenArea * ScreenArea::Add( const ScreenArea * addition ) const
{

    ScreenArea * rslt;

    rslt = new ScreenArea( *this );

    for ( int i = 0 ; i < addition->RectangleCount; i++ )
    {
        rslt->AddSubstracting( addition->Rectangles[ i ], this, 0 );
    }
    return rslt;
}

void ScreenArea::bounding(int & minx,int & miny,int & w,int & h) const
{
    int x0, y0, x1, y1;
    x0 = 0x3fffffff;
    y0 = 0x3fffffff;
    x1 = -0x3fffffff;
    y1 = -0x3fffffff;
    for(int i = 0; i < RectangleCount; ++i) {
	if (x0 > Rectangles[i].X) x0 = Rectangles[i].X;
	if (y0 > Rectangles[i].Y) y0 = Rectangles[i].Y;
	if (x1 < Rectangles[i].X + Rectangles[i].W - 1) x1 = Rectangles[i].X + Rectangles[i].W - 1;
	if (y1 < Rectangles[i].Y + Rectangles[i].H - 1) x1 = Rectangles[i].Y + Rectangles[i].H - 1;
    }
    minx = x0;
    miny = y0;
    w = x1 - x0 + 1;
    h = y1 - y0 + 1;
}

ScreenArea::ScreenArea()
{
    RectangleCount = 0;
    Rectangles = 0;
}

ScreenArea::ScreenArea(int x,int y,int w,int h)
{
    RectangleCount = 0;
    Rectangles = 0;
    if (x < 0 ) {
        w+=x;
        x = 0;
        if (w <= 0) { return; }
    }
    
    if (y < 0) {
        h+=y;
        y = 0;
        if (h <= 0) { return; }
    }
    
    int rid = NewRectangle();
    
    Rectangles[rid].X = x;
    Rectangles[rid].Y = y;
    Rectangles[rid].W = w;
    Rectangles[rid].H = h;
    
}

void ScreenArea::Add(int x,int y,int w,int h)
{
    int rid = NewRectangle();
    
    Rectangles[rid].X = x;
    Rectangles[rid].Y = y;
    Rectangles[rid].W = w;
    Rectangles[rid].H = h;
}

ScreenArea::ScreenArea( const ScreenArea & cpy )
{
    RectangleCount = cpy.RectangleCount;
    if ( RectangleCount )
    {
        int AllocNb = RectangleStorageFor( RectangleCount );

        Rectangles = ( Rectangle* ) malloc( sizeof( Rectangle ) * AllocNb );
        memcpy( Rectangles, cpy.Rectangles, RectangleCount * sizeof( Rectangle ) );
    }
    else
    {
        Rectangles = 0;
    }
}

ScreenArea::~ScreenArea()
{
    if ( Rectangles )
    {
        free( Rectangles );
    }
}

int ScreenArea::NewRectangle()
{
    int curStorage = RectangleStorageFor( RectangleCount );
    int newStorage = RectangleStorageFor( RectangleCount + 1 );

    if ( curStorage != newStorage )
    {
        Rectangles = ( Rectangle* ) realloc( ( void* ) Rectangles, sizeof( Rectangle ) * newStorage );
    }
    return RectangleCount++;
}

void ScreenArea::DelRectangle( int id )
{
    int curStorage = RectangleStorageFor( RectangleCount );
    int newStorage = RectangleStorageFor( RectangleCount - 1 );

    Rectangles[ id ] = Rectangles[ RectangleCount - 1 ];
    RectangleCount--;

    if ( curStorage != newStorage )
    {
        Rectangles = ( Rectangle* ) realloc( ( void* ) Rectangles, sizeof( Rectangle ) * newStorage );
    }
}

void ScreenArea::Debug() const
{
    for ( int i = 0 ;i < RectangleCount;i++ )
    {
        printf("%dx%d@(%d,%d) ",(int)Rectangles[i].W,(int)Rectangles[i].H,(int)Rectangles[i].X,(int)Rectangles[i].Y);
    }
    printf("\n");
}

bool ScreenArea::IsEmpty() const
{
    return RectangleCount == 0;
}

bool ScreenArea::Contact( const Rectangle& r ) const
{
    Rectangle result;

    for ( int i = 0 ;i < RectangleCount;i++ )
    {
        if ( RectangleIntersect( &r, Rectangles + i, &result ) )
        {
            return true;
        }
    }
    return false;
}

bool ScreenArea::Contact( const ScreenArea & r ) const
{
    Rectangle result;
    for ( int i = 0 ;i < RectangleCount;i++ )
    {
        for ( int j = 0 ;j < r.RectangleCount; j++ )
        {
            if ( RectangleIntersect( Rectangles + i, r.Rectangles + j, & result ) )
            {
                return true;
            }
        }
    }
    return false;
}

