#include <SDL.h>
#include <SDL_image.h>

#include "Graphs.h"

#include "jpeglib.h"
#include <setjmp.h>

#include <string>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

typedef struct surface_error_handler {
    struct jpeg_error_mgr pub;
    jmp_buf error_jump;
} surface_error_handler;

static void surface_error_exit (j_common_ptr cinfo)
{
    /* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
    surface_error_handler * myerr = (surface_error_handler*) cinfo->err;

    /* Always display the message. */
    /* We could postpone this until after returning, if we chose. */
    (*cinfo->err->output_message) (cinfo);

    /* Return control to the setjmp point */
    longjmp(myerr->error_jump, 1);
}

typedef struct surface_jpeg_compress_struct {
    jpeg_compress_struct jpg;
    Surface * RelatedTo;
    std::string * destination;
    JOCTET buffer[4096];
    
} surface_jpeg_compress_struct;

static void jpg_enc_init_destination(j_compress_ptr cinfo)
{
    surface_jpeg_compress_struct * s = (surface_jpeg_compress_struct*) cinfo;

    s->jpg.dest->next_output_byte = s->buffer;
    s->jpg.dest->free_in_buffer = 4096;
}

static int jpg_enc_empty_output_buffer(j_compress_ptr cinfo)
{
    surface_jpeg_compress_struct * s = (surface_jpeg_compress_struct*) cinfo;
    (*s->destination)+=std::string((const char*)s->buffer,4096);
    s->jpg.dest->next_output_byte = s->buffer;
    s->jpg.dest->free_in_buffer = 4096;
    return TRUE;
}

static void jpg_enc_term_destination(j_compress_ptr cinfo)
{
    surface_jpeg_compress_struct * s = (surface_jpeg_compress_struct*) cinfo;
    
    (*s->destination)+=std::string((const char*)s->buffer,4096 - s->jpg.dest->free_in_buffer);
}

typedef struct surface_jpeg_decompress_struct {
    jpeg_decompress_struct jpg;
    Surface * RelatedTo;
    const std::string * source;
    int skip;
    jmp_buf error_jump;
} surface_jpeg_decompress_struct;

static void jpg_dec_init_source(j_decompress_ptr cinfo)
{
    surface_jpeg_decompress_struct * s = (surface_jpeg_decompress_struct*) cinfo;
    s->jpg.src->next_input_byte = (JOCTET*) s->source->data() + s->skip;
    s->jpg.src->bytes_in_buffer = s->source->size() - s->skip;
}

static int jpg_dec_fill_input_buffer(j_decompress_ptr cinfo)
{
    surface_jpeg_decompress_struct * s = (surface_jpeg_decompress_struct*) cinfo;
    s->jpg.src->bytes_in_buffer = 0;
    return TRUE;
}

static void jpg_dec_skip_input_data(j_decompress_ptr cinfo, long num_bytes)
{
    surface_jpeg_decompress_struct * s = (surface_jpeg_decompress_struct*) cinfo;
    if (num_bytes <= 0) return ;
    if (num_bytes > (long)s->jpg.src->bytes_in_buffer) {
	s->jpg.src->bytes_in_buffer = 0;
	return ;
    }
    s->jpg.src->bytes_in_buffer -= num_bytes;
    s->jpg.src->next_input_byte += num_bytes;
}

static void jpg_dec_term_source(j_decompress_ptr cinfo)
{
}

Surface::Surface()
{
    Surf = 0;
    DisplayTuned = false;
    
}

void Surface::LooseSurface()
{
    if (Surf) {
	SDL_FreeSurface(Surf);
	Surf = 0;
    }
}

Surface::~Surface()
{
    LooseSurface();
}

bool Surface::IsPaletted() const
{
    return Surf && Surf->format->palette;
}

int Surface::MapRGB(int r,int g,int b) const
{
    if (!Surf) return 0;
    return SDL_MapRGB(Surf->format,r,g,b);
}

Uint32 Surface::mapColor(const Color & c) const
{
    if (!Surf) return 0;
    return SDL_MapRGBA(Surf->format,c.r(),c.g(),c.b(),c.a());
}

int Surface::GetWidth() const
{
    return (Surf ? Surf->w : 0);
}

int Surface::GetHeight() const
{
    return (Surf ? Surf->h : 0);
}

bool Surface::HasTransparency() const
{
    return Format!=Opaque;
}

SDL_Surface * Surface::GetSDLSurface() const
{
    return Surf;
}

void Surface::assign(SDL_Surface * s)
{
    LooseSurface();
    Format = Opaque;
    DisplayTuned = true;
    Surf = s;
}

void append_uint32(std::string & rslt,Uint32 v)
{
    unsigned char buffer[4];
    buffer[0] = v & 255;
    buffer[1] = (v >> 8) & 255;
    buffer[2] = (v >> 16) & 255;
    buffer[3] = (v >> 24) & 255;
    rslt += std::string((const char*)buffer,4);
    //jsrlog("appended %#x => newsize is %d\n",v,rslt.length());
}

Uint32 get_uint32(const std::string & rslt,int & idx)
{
    Uint32 buffer[4];
    Uint32 r;
    for(int i = 0 ; i < 4; ++i) {
	buffer[i] = ((unsigned char)rslt[i+idx]);
    }
    idx += 4;
    r = buffer[0] | (buffer[1] << 8) | (buffer[2] << 16) | (buffer[3] << 24);
    //jsrlog("@idx=%d, got %#x\n",idx,r);
    return r;
}

void Surface::encode(std::string & rslt)
{
    if (Surf && Surf->w && Surf->h) {
	append_uint32(rslt,1);
	
	struct surface_jpeg_compress_struct compress_st;
	struct jpeg_destination_mgr mgr;
	
	compress_st.destination = &rslt;
	JSAMPROW row_pointer[1];
	
	mgr.init_destination = &jpg_enc_init_destination;
	mgr.empty_output_buffer = &jpg_enc_empty_output_buffer;
	mgr.term_destination= &jpg_enc_term_destination;
	
	struct surface_error_handler jerr;
	compress_st.jpg.err = jpeg_std_error(&jerr.pub);
	jerr.pub.error_exit = &surface_error_exit;
	
	if (setjmp(jerr.error_jump) != 0) {
	    jsrlog("jpeg compression error\n");
	    if (setjmp(jerr.error_jump) == 0) {
		jpeg_finish_compress(&compress_st.jpg);
	    }
	    if (setjmp(jerr.error_jump) == 0) {
		jpeg_destroy_compress(&compress_st.jpg);
	    }
	    SDL_UnlockSurface(Surf);
	    // FIXME : clear rslt ?
	    rslt.clear();
	    return;
	}
	
	jpeg_create_compress(&compress_st.jpg);
	SDL_LockSurface(Surf);
	
	
	
	compress_st.jpg.dest = &mgr;
	compress_st.jpg.image_width = Surf->w; 	/* image width and height, in pixels */
	compress_st.jpg.image_height = Surf->h;
	compress_st.jpg.input_components = 3;		/* # of color components per pixel */
	compress_st.jpg.in_color_space = JCS_RGB; 	/* colorspace of input image */
	
	jpeg_set_defaults(&compress_st.jpg);
	jpeg_set_quality(&compress_st.jpg, 90, TRUE /* limit to baseline-JPEG values */);
	jpeg_start_compress(&compress_st.jpg, TRUE);
	while (compress_st.jpg.next_scanline < compress_st.jpg.image_height) {
	    /* Here p is the address to the pixel we want to retrieve */
	    Uint8 *p = (Uint8 *)Surf->pixels + compress_st.jpg.next_scanline * Surf->pitch ;
	    
	    /* jpeg_write_scanlines expects an array of pointers to scanlines.
	     * Here the array is only one element long, but you could pass
	     * more than one scanline at a time if that's more convenient.
	     */
	    row_pointer[0] = p;
	    (void) jpeg_write_scanlines(&compress_st.jpg, row_pointer, 1);
	}
	
	jpeg_finish_compress(&compress_st.jpg);
	jpeg_destroy_compress(&compress_st.jpg);
	SDL_UnlockSurface(Surf);
	//fclose(outfile);
    } else {
	append_uint32(rslt,0);
    }
}

void Surface::decode(const std::string & rslt)
{
    int idx;
    idx = 0;
    LooseSurface();
    
    if (!get_uint32(rslt,idx)) {
	return;
    }
    
    struct surface_jpeg_decompress_struct decompress_st;
    struct jpeg_source_mgr mgr;
    
    struct surface_error_handler jerr;
    decompress_st.jpg.err = jpeg_std_error(&jerr.pub);
    jerr.pub.error_exit = &surface_error_exit;
    
    
    if (setjmp(jerr.error_jump) != 0) {
	if (setjmp(jerr.error_jump) == 0) {
	    jpeg_finish_decompress(&decompress_st.jpg);
	}
	if (setjmp(jerr.error_jump) == 0) {
	    jpeg_destroy_decompress(&decompress_st.jpg);
	}
	if (Surf) {
	    SDL_UnlockSurface(Surf);
	    LooseSurface();
	}
	return;
    }
    
    jpeg_create_decompress(&decompress_st.jpg);
    
    mgr.init_source = &jpg_dec_init_source;
    mgr.fill_input_buffer = &jpg_dec_fill_input_buffer;
    mgr.skip_input_data = &jpg_dec_skip_input_data;
    mgr.resync_to_restart = &jpeg_resync_to_restart;
    mgr.term_source = &jpg_dec_term_source;
    decompress_st.source = &rslt;
    decompress_st.skip = idx;
    decompress_st.jpg.src = &mgr;
    
    jpeg_read_header(&decompress_st.jpg, TRUE);
    
    jpeg_start_decompress(&decompress_st.jpg);
    
    Create(	decompress_st.jpg.output_width,
    		decompress_st.jpg.output_height,
    		Opaque,false,0);
    SDL_LockSurface(Surf);

    while (decompress_st.jpg.output_scanline < decompress_st.jpg.output_height) 
    {
	int n = decompress_st.jpg.output_height - decompress_st.jpg.output_scanline;
	JSAMPROW row_pointer[n];
	
	for(int i = 0; i < n; ++i) {
	    Uint8 *p = (Uint8 *)Surf->pixels + (decompress_st.jpg.output_scanline + i) * Surf->pitch ;
	    row_pointer[i] = p;
	}
	jpeg_read_scanlines(&decompress_st.jpg,row_pointer, n);
    }
    jpeg_finish_decompress(&decompress_st.jpg);
    jpeg_destroy_decompress(&decompress_st.jpg);
    SDL_UnlockSurface(Surf);
}

void Surface::Create(int width,int height,FormatFlag format, bool displayTune, int ckey)
{
    LooseSurface();
    
    Format = format;
    DisplayTuned = false;
    
    Uint32 rmask, gmask, bmask, amask;

    /* SDL interprets each pixel as a 32-bit number, so our masks must depend
       on the endianness (byte order) of the machine */
    #if SDL_BYTEORDER == SDL_BIG_ENDIAN
	rmask = 0x0000ff00;
	gmask = 0x00ff0000;
	bmask = 0xff000000;
	amask = 0x000000ff;
    #else
	rmask = 0x00ff0000;
	gmask = 0x0000ff00;
	bmask = 0x000000ff;
	amask = 0xff000000;
    #endif
    if (format != Alpha) { amask = 0; }
    Surf = SDL_CreateRGBSurface(SDL_SWSURFACE | 
			    (format == ColorKey ? SDL_SRCCOLORKEY : 0) |
			    (format == Alpha ? SDL_SRCALPHA : 0),
			    width,
			    height,
			    (format == Alpha ? 32 : 24),
			    rmask, gmask, bmask, amask);
    if ((format == ColorKey) && (ckey != -1)) {
	SetColorKey(ckey);
    }
    
    
    if (displayTune) {
	DisplayTune();
    }
}

void Surface::clear(Uint32 c)
{
    memset(Surf->pixels,0,Surf->w*Surf->h*(Format == Alpha ? 4 : 3));
}

bool Surface::LoadImage(const char * name, FormatFlag format, bool displayTune, int ckey)
{
    SDL_Surface * tmp = IMG_Load( name );
    if ( !tmp )
    {
	fprintf(stderr,"IMG_Load %s : %s\n",name, IMG_GetError());
        return false;
    }
    
    
    jsrlog( "image size : %dx%d\n", tmp->w, tmp->h );
    jsrlog( "format : %d\n", tmp->format->BitsPerPixel );
    
    LooseSurface();
    DisplayTuned = false;
    Format = format;
    
    Surf = tmp;
    if ((format == ColorKey) && (ckey != -1)) {
	SetColorKey(ckey);
    }
    
    if (displayTune) {
	DisplayTune();
    }
    return true;
}

void Surface::SetTransparency(int l)
{
    if (DisplayTuned || !Surf) return;
    if (l == 255 ) {
	SDL_SetAlpha(Surf, 0, 255);
    } else {
	SDL_SetAlpha(Surf, SDL_SRCALPHA, l);
    }
}

void Surface::SetColorKey(int ckey)
{
    if (DisplayTuned || !Surf) return;
    
    if (IsPaletted()) {
	SDL_SetColorKey(Surf, SDL_SRCCOLORKEY | SDL_RLEACCEL, ckey);
    } else {
	SDL_SetColorKey(Surf, SDL_SRCCOLORKEY | SDL_RLEACCEL,
	    MapRGB( (ckey & 0x00ff0000),
		    (ckey & 0x0000ff00),
		    (ckey & 0x000000ff)));
    }
}

void Surface::DisplayTune()
{
    if (!Surf) return;
    if (DisplayTuned) return;
    
    SDL_Surface * newSurf;
    if (Format == Alpha) {
	newSurf = SDL_DisplayFormatAlpha(Surf);
    } else {
	newSurf = SDL_DisplayFormat(Surf);
    }
    DisplayTuned = true;
    SDL_FreeSurface(Surf);
    Surf = newSurf;
    return;
}


