/***************************************************************************
*   Copyright (C) 2004 by ludovic p                                       *
*   pludov@nnx.com                                                        *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <stdlib.h>


#include "jsreflect.h"

#include "jsdbgapi.h"
#include "jsscript.h"
#include "i18n.h"

using namespace JSReflection;

FunctionInfo * Context::PublicFunctions = 0;

static void script_error_reporter( JSContext *cx, const char *message, JSErrorReport *report )
{
    if ( report == NULL )
    {
        fprintf( stderr, "jssh: %s\n", message );
    }
    else
    {
        fprintf( stderr, "%s: line %i: error: %s\n",
                 report->filename, report->lineno, message );
    }
}

static JSTrapStatus script_interrupt_callback(JSContext *cx, JSScript *script,jsbytecode *pc,
                                  jsval *rval, void *closure)
{
    //fprintf(stderr,"JS pass at %s line %d\n",(script->filename?script->filename:"<unknown>"),script->lineno);
    
    return JSTRAP_CONTINUE;
}

static JSBool script_print( JSContext * cx, JSObject * obj, uintN argc, jsval * argv, jsval * rval )
{
    for ( unsigned  i = 0;i < argc; i++ )
    {
        char * tmp;
        if (JS_ConvertArguments(cx, 1 , argv + i , "s", &tmp)) {
	    printf("%s",tmp);
        }
    }
    *rval = JSVAL_VOID;
    return JS_TRUE;
}

static JSBool script_debug( JSContext * cx, JSObject * obj, uintN argc, jsval * argv, jsval * rval )
{
    if (jsr_debug_enabled()) {
	for ( unsigned  i = 0;i < argc; i++ )
	{
	    char * tmp;
	    if (JS_ConvertArguments(cx, 1 , argv + i , "s", &tmp)) {
		printf("%s",tmp);
	    }
	}
    }
    *rval = JSVAL_VOID;
    return JS_TRUE;
}

static JSBool script_i18n( JSContext * cx, JSObject * obj, uintN argc, jsval * argv, jsval * rval )
{
    std::string * strargv[argc];
    std::string result;
    JSReflection::Context * c = JSReflection::Context::Find(cx);
    
    for(int i = 0 ; i < argc; ++i) { 
	char * tmp;
	JS_String str;
	if (c->ValueFromJS(argv + i ,&str, "%s", 0)) {
	    std::string t = str;
	    strargv[i] = new std::string(t);
	} else {
	    strargv[i] = new std::string("");
	}
    }
    result = i18n(*strargv[0],argc - 1,strargv + 1);
    for(int i = 0; i < argc; ++i) {
	delete(strargv[i]);
    }
    JS_String str;
    str = result;
    c->ValueToJS(&str,rval,"%s",0);
    return JS_TRUE;
}


static JSFunctionSpec default_functions[] =
    {
	{"print", script_print, 1, 0, 0},
	{"debug", script_debug, 1, 0, 0},
	{"i18n",  script_i18n,  1, 0, 0},
	{0, 0, 0, 0, 0}
    };


Context * Context::Find( JSContext * c )
{
    void * priv;
    priv = JS_GetContextPrivate( c );
    if ( !priv )
    {
        return 0;
    }
    return ( Context* ) priv;
}

Context::Context( JSRuntime * js_runtime, int stack_size )
{
    JSBool ok;
    JSContext * js_context;
    JSObject * js_global;
    FunctionInfo * globalf;

    First = 0;
    Last = 0;
    runtime = js_runtime;
    
    /* create a context and associate it with the JS run time */
    js_context = JS_NewContext( js_runtime, stack_size );

    /* if cx does not have a value, end the program here */
    if ( js_context == NULL )
    {
        fprintf( stderr, "JS_NewContext failed\n" );
        exit( 1 );
    }

    JS_SetErrorReporter( js_context, script_error_reporter );
    JS_SetContextPrivate( js_context, this );
    JS_SetOptions(js_context, JSOPTION_STRICT);
    JS_SetInterrupt( js_runtime, script_interrupt_callback,0);
    
    
    /* create the global object here */
    js_global = JS_NewObject( js_context, 0, NULL, NULL );
    global = new Object(this);
    global->Counterpart = js_global;
    
    /* initialize the built-in JS objects and the global object */
    ok = JS_InitStandardClasses( js_context, js_global );
    if ( ok != JS_TRUE )
    {
        fprintf( stderr, "Failed to initialize standard classes\n" );
        exit( 1 );
    }

    ok = JS_DefineFunctions( js_context, js_global, default_functions );
    if ( ok != JS_TRUE )
    {
        fprintf( stderr, "Failed to register specific classes\n" );
        exit( 1 );
    }

    globalf = PublicFunctions;
    while ( globalf )
    {
        JS_DefineFunction( js_context, js_global, globalf->Name, globalf->Func, globalf->NArgs, 0 );
        globalf = globalf->Next;
    }

    int classnb;

    
    JSGlobalContext = js_context;
    JSGlobalObject = js_global;
    JS_AddRoot( JSGlobalContext, &JSGlobalObject );

    classnb = ClassInfo::GetClassCount();
    if ( classnb )
    {
        Models = ( JSObject** ) malloc( sizeof( JSObject* ) * classnb );
        for ( int i = 0 ; i < classnb; i++ )
        {
            Models[ i ] = 0;
            JS_AddRoot( JSGlobalContext, Models + i );
        }
    }
    else
    {
        Models = 0;
    }
}

Context::~Context()
{
    while ( First )
    {
        First->SetCounterpart( 0 );
        First = First->Next;
    }

    if ( Models )
    {
        int classnb = ClassInfo::GetClassCount();

        for ( int i = 0; i < classnb; i++ )
        {
            JS_RemoveRoot( JSGlobalContext, Models + i );
        }

        free( Models );
    }

    JS_RemoveRoot( JSGlobalContext, &JSGlobalObject );
    JSGlobalObject = 0;
    JS_DestroyContext( JSGlobalContext );
    JSGlobalContext = 0;
}

static void debugroot(const char *name, void *rp, void *data)
{
    printf("%s\n",name);
}

void Context::dumpRoots() {
    //JS_DumpNamedRoots(runtime,&debugroot,0);
}

JSContext * Context::GetJSContext() const
{
    return JSGlobalContext;
}

Object * Context::GetGlobal() const
{
    return global;
}

bool Context::ExecuteScript( const char *scripttext, Object * root, jsval * retvalue,const JSSourceInfo & jsfi )
{
    if (root) {
	return ExecuteScript(scripttext,root->GetJSObject(),retvalue,jsfi);
    } else {
	return ExecuteScript(scripttext,(JSObject*)0,retvalue,jsfi);
    }
}

bool Context::ExecuteScript( const char *scripttext, JSObject * root, jsval * retvalue,const JSSourceInfo & jsfi )
{
    JSScript * script;
    JSBool status;
    jsval _tmp_result;
    jsval * result = (retvalue ? retvalue : &_tmp_result);
	
    jsrlog("Context::ExecuteScript %s:%d\n",jsfi.getName().c_str(),jsfi.getLine());
    
    JSObject * realRoot;
    if (root) {
	realRoot = root;
    } else {
	realRoot = JSGlobalObject;
    }
	
    script = JS_CompileScript( JSGlobalContext, realRoot, scripttext, strlen(scripttext), jsfi.getName().c_str(),jsfi.getLine() );
    
    
    if ( script == NULL )
    {
        fprintf(stderr,"failed to compile <internal script>\n");
        return false;
    }

    status = JS_ExecuteScript( JSGlobalContext, realRoot, script, result );
    JS_DestroyScript( JSGlobalContext, script );

    if ( status == JS_TRUE )
    {
        return true;
    }
    else
    {
        fprintf(stderr, "execution of <internal script> failed\n");
        return false;
    }
}

bool Context::ExecuteScriptFile( const char *name, Object * root )
{
    FILE * f;
    int size, readed;
    char * buffer;

    JSScript * script;
    JSBool status;
    jsval result;

    f = fopen( name, "r" );
    if ( f == NULL )
    {
        perror( name );
        return false;
    }
    fseek( f, 0L, SEEK_END );
    size = ftell( f );

    fseek( f, 0L, SEEK_SET );
    buffer = ( char* ) malloc( size );
    readed = fread( buffer, 1, size, f );
    fclose( f );

    JSObject * realRoot;
    if (root) {
	realRoot = root->GetJSObject(true);
    } else {
	realRoot = JSGlobalObject;
    }

    jsrlog("Context::ExecuteScriptFile %s\n",name);
    
    script = JS_CompileScript( JSGlobalContext, realRoot, buffer, readed, name, 1 );
    
    
    if ( script == NULL )
    {
        fprintf(stderr, "failed to compile %s\n", name );
        return false;
    }

    status = JS_ExecuteScript( JSGlobalContext, realRoot, script, &result );
    JS_DestroyScript( JSGlobalContext, script );

    if ( status == JS_TRUE )
    {
        return true;
    }
    else
    {
        fprintf(stderr, "execution of %s failed\n", name );
        return false;
    }
}


void Context::DumpPredefined()
{
    FunctionInfo * act;
    act = PublicFunctions;
    while(act) {
	act->Debug();
	act = act->Next;
    }
    ClassInfo * cl;
    cl = ClassInfo::First;
    while(cl) {
	cl->Debug();
	cl = cl->Next;
    }
}

Object * Context::GetAssociatedObject( JSObject * o ) const
{
    Object * c_o;
    c_o = ( Object* ) JS_GetPrivate( JSGlobalContext, o );
    if (c_o && c_o->isDebugZombi()) {
	// Zombi doesn't count !
	return 0;
    }
    return c_o;
}

bool Context::ValueToJS( void * in, jsval * out, const char * format, ClassInfo * specifier )
{
    if ( format[ 0 ] != '%' )
    {
        jsrlog("invalid format : %s\n",format);
	return false;
    }
    while ( format[ 0 ] )
    {
        switch ( format[ 0 ] )
        {
            case '!':
            break;
            case '+':
            break;
            case 'i':
            {
                int * vin = ( int* ) in;
                if ( !INT_FITS_IN_JSVAL( *vin ) )
                {
                    // FIXME : throw overflow
		     jsrlog("integer overflow : %d\n",*vin);
                    return false;
                }
                *out = INT_TO_JSVAL( *vin );
            }
            return true;

            case 'b':
            {
                bool * vin = ( bool* ) in;
                *out = BOOLEAN_TO_JSVAL( *vin );
            }
	    return true;
            

            case 's':
            {
                JS_String * string = ( JS_String* ) in;
                char * data = ( char* ) string->data();
                *out = STRING_TO_JSVAL( JS_NewString( JSGlobalContext, data, strlen( data ) ) );
                string->SetFree( 0 );
            }
            return true;

            case 'x':
            {
                Object * realvalue;
                if ( !*( void** ) in )
                {
                    realvalue = 0;
                }
                else
                {
                    if ( !specifier )
                    {
                        realvalue = * ( ( Object** ) in );
                    }
                    else
                    {
                        realvalue = specifier->CastFromSubClass( *( void** ) in );
                    }
                }
                if ( realvalue )
                {
                    *out = OBJECT_TO_JSVAL( realvalue->GetJSObject() );
                }
                else
                {
                    *out = JSVAL_NULL;
                }
            }
            return true;
        }
        format++;
    }
    jsrlog("bad format ???\n");
    return false;
}

bool Context::ValueFromJSObject( JSObject * o, void * out, const char * format, ClassInfo * specifier )
{
    jsval v;
    v = OBJECT_TO_JSVAL( o );
    return ValueFromJS( &v, out, format, specifier );
}

bool Context::ValueFromJS( jsval * in, void * out, const char * format, ClassInfo * specifier )
{
    bool notnull;
    bool positive;

    if ( format[ 0 ] != '%' )
    {
	jsrlog("bad format : %s\n",format);
        return false;
    }

    format++;

    notnull = false;
    positive = false;
    while ( format[ 0 ] )
    {
        switch ( format[ 0 ] )
        {
            case '!':
            notnull = true;
            break;
            case '+':
            positive = true;
            break;
            case 'x':     // Object
            if ( ( !notnull ) && ( JSVAL_IS_NULL( *in ) || JSVAL_IS_VOID( *in ) ) )
            {
                ( *( Object** ) out ) = 0;
                return true;
            }
            if ( JSVAL_IS_OBJECT( *in ) )
            {
                JSObject * v = JSVAL_TO_OBJECT( *in );
                Object* o = GetAssociatedObject( v );
                if ( !o )
                {
		    ( *( Object** ) out ) = 0;
		    if (notnull) jsrlog("unexpected null : wrong object ?\n");
		    return !notnull;
                }
                if ( specifier )
                {
                    if ( o->GetClassId() ->SubClassOf( specifier ) )
                    {
                        ( *( void** ) out ) = specifier->CastToSubClass( o );
                    }
                    else
                    {
			jsrlog("dynamic type check failed\n");
                       return false;
                    }
                }
                else
                {
                    ( *( Object** ) out ) = o;
                }
                return true;
            }
	    jsrlog("object expected\n");
            return false;

            case 's':     // String
            {
                char * tmp;
                JS_String * vout = ( JS_String* ) out;
                if (JS_ConvertArguments(JSGlobalContext, 1 , in , "s", &tmp)) {
                    vout->Copy( tmp,strlen(tmp));
                    return true;
                } else {
		     jsrlog("can't get string\n");
                    return false;
                }
            }
            return false;

            case 'b':
            if ( JSVAL_IS_BOOLEAN( *in ) )
            {
                bool * vout = ( bool* ) out;
                *vout = (JSVAL_TO_BOOLEAN( *in ) != 0);
                return true;
            }
	    jsrlog("not a boolean\n");
            return false;

            case 'i':     // Integer
            if ( JSVAL_IS_INT( *in ) )
            {
                int * vout = ( int* ) out;
                *vout = JSVAL_TO_INT( *in );
                if ( positive && ( ( *vout ) < 0 ) )
                {
		     jsrlog("positive number expected (got %d)\n",(*vout));
                    return false;
                }
                if ( notnull && ( ( *vout ) == 0 ) )
                {
		     jsrlog("notnull number expected\n");
                    return false;
                }
                return true;
            }
	    if (JSVAL_IS_DOUBLE( * in ))
	    {
		jsdouble  rslt;
		if (!JS_ValueToNumber(JSGlobalContext, *in, &rslt)) {
		    jsrlog("unable to convert to number\n");
		    return false;
		}
		int * vout = ( int* ) out;
		*vout = (int) (rslt);
		if ( positive && ( ( *vout ) < 0 ) )
                {
		    jsrlog("positive number expected (got %d)\n",(*vout));
                    
                    return false;
                }
                if ( notnull && ( ( *vout ) == 0 ) )
                {
                    jsrlog("notnull number expected\n");
                    return false;
                }
                return true;
	    }
	    jsrlog("unable to convert to number\n");
                    
            return false;
        }
        format ++;
    };
    jsrlog("wrong format ???\n");
    return false;
}


